package it.polito.bigdata.spark;
import org.apache.spark.api.java.*;

import scala.Tuple2;

import java.util.List;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.spark.SparkConf;

public class SparkDriver {

    public static void main(String[] args) {
        
		Logger.getLogger("org").setLevel(Level.OFF);
		Logger.getLogger("akka").setLevel(Level.OFF);

        String jobContractsPath = "sample_data/JobContracts.txt";
        String jobOffersPath = "sample_data/JobOffers.txt";
        String jobPostingsPath = "sample_data/JobPostings.txt";
        
        String outputPath1 = "outSpark1/";
        String outputPath2 = "outSpark2/";

        // Initialize SparkContext
        SparkConf conf = new SparkConf().setAppName("Exam 20240620_Spark").setMaster("local");
		// Create a Spark Context object
		JavaSparkContext sc = new JavaSparkContext(conf);

        // ########################################
        // PART 1
        // ########################################
        JavaRDD<String> jobPostingsRDD = sc.textFile( jobPostingsPath );    // JobID,Title,Country
        JavaRDD<String> jobOffersRDD = sc.textFile( jobOffersPath);         // OfferID,JobID,Salary,Status,SSN
        JavaRDD<String> jobContractsRDD = sc.textFile( jobContractsPath );  // ContractID,OfferID,ContractDate,ContractType

        // Filter accepted job offers and extract (JobID,(Salary, OfferId))
        // OfferId is helpful for solving the second part.
        // To avoid repeting the same join in the second part, we already retrieve the OfferId here.

        JavaPairRDD<String, Tuple2<Double, String>> acceptedOffers = jobOffersRDD
            // OfferID,JobID,Salary,Status,SSN
            .filter(line -> line.contains(",Accepted,"))
            .mapToPair(line -> {
                String[] fields = line.split(",");
                String offerId = fields[0];
                String jobID = fields[1];
                double salary = Double.parseDouble(fields[2]);
                return new Tuple2<>(jobID, new Tuple2<>(salary,offerId));
            });

        // Extract (JobID, (Country, Title)) from job postings
        // Title is helpful for solving the second part.
        // To avoid repeting the same join in the second part, we already retrieve the title here.
        JavaPairRDD<String, Tuple2<String,String>> jobIDCountry = jobPostingsRDD
            // JobID,Title,Country
            .mapToPair(line -> {
                String[] fields = line.split(",");
                String jobID = fields[0];
                String title = fields[1];
                String country = fields[2];
                return new Tuple2<>(jobID, new Tuple2<>(country,title));
            });

        // Join accepted offers with job postings to get (JobID, ((Salary, OfferId), (Country, Title)))
        JavaPairRDD<String, Tuple2<Tuple2<Double, String>, Tuple2<String,String>>> offersWithCountry = 
                            acceptedOffers.join(jobIDCountry).cache();

        // Map to (Country, (Salary, 1))
        JavaPairRDD<String, Tuple2<Double, Integer>> countrySalaryCount = offersWithCountry
            .mapToPair(tuple -> new Tuple2<>(tuple._2()._2()._1(), new Tuple2<>(tuple._2()._1()._1(), 1)));

        // Reduce by key to get (Country, (TotalSalary, Count))
        JavaPairRDD<String, Tuple2<Double, Integer>> countryTotalSalaryCount = countrySalaryCount
            .reduceByKey((a, b) -> new Tuple2<>(a._1() + b._1(), a._2() + b._2()));

        // Map to (Country, AverageSalary)
        JavaPairRDD<String, Double> countryAverageSalary = countryTotalSalaryCount
            .mapValues(tuple -> tuple._1() / tuple._2());

        // Select only the top N countries with the highest average salary
        List<Tuple2<String, Double>> topCountries = countryAverageSalary
            .mapToPair(tuple -> new Tuple2<>(tuple._2(), tuple._1()))
            .sortByKey(false)
            .mapToPair(tuple -> new Tuple2<>(tuple._2(), tuple._1()))
            .take(3);

        // Convert the top 3 countries to an RDD and save the result
        JavaRDD<Tuple2<String, Double>> topCountriesRDD = sc.parallelize(topCountries);
        topCountriesRDD.saveAsTextFile(outputPath1);
        
        
        // ########################################
        // PART 2
        // ########################################

        // Map offersWithCountry to
        // (OfferID, (Country, Title))
        JavaPairRDD<String, Tuple2<String,String>> offerTitleCountry = 
                offersWithCountry.mapToPair(pair -> {
                    String offerId = pair._2()._1()._2();
                    String country = pair._2()._2()._1();
                    String title = pair._2()._2()._2();

                     return new Tuple2<>(offerId, new Tuple2<>(title, country));
                });



        // Map contracts to (OfferID, null)
        JavaPairRDD<String, String> offerIdContract = jobContractsRDD
            .mapToPair(line -> {
                // ContractID,OfferID,ContractDate,ContractType
                String[] fields = line.split(",");
                String offerID = fields[1];
                return new Tuple2<>(offerID, null);
            });

        // Join offerTitleCountry with offerIdContract -> (offerId, ((Title, Country), null))
        // map to ((title, country), +1)
        // and apply reduceByKey to count the number of contracts for each title in each country
        // ((title, country), numContracts)
        JavaPairRDD<Tuple2<String, String>, Integer> titleCountryNumContracts = 
        offerTitleCountry.join(offerIdContract)
            .mapToPair(pair -> {
                Tuple2<String,String> countryTitle = pair._2()._1();
                return new Tuple2<>(countryTitle, 1);
            })
            .reduceByKey((v1, v2) -> v1+v2).cache();


        // Map to (Country, numContracts) and compute the maximum for each country
        JavaPairRDD<String, Integer> countryMaxNumContracts = titleCountryNumContracts
            .mapToPair(pair -> {
                String country =  pair._1()._2();
                Integer numContracts = pair._2();

                return new Tuple2<>(country, numContracts);
            })
        .reduceByKey((v1, v2) -> v1 > v2 ? v1 : v2);
        

        // Map countryMaxNumContracts to 
        // ( (Country, maxNumContracts), null )
        JavaPairRDD<Tuple2<String, Integer>, String> countryMaxNull = countryMaxNumContracts
                    .mapToPair(pair -> new Tuple2<>(pair, null));


        // Map Join titleCountryNumContracts to ((country, numContracts), title)
        JavaPairRDD<Tuple2<String, Integer>, String> countryNumContractsTitle = 
            titleCountryNumContracts.mapToPair(pair -> {
                String title = pair._1()._1();
                String country = pair._1()._2();
                Integer numContracts = pair._2();

                return new Tuple2<>(new Tuple2<>(country,numContracts), title);
            });

        // Join countryNumContractsTitle with countryMaxNull
        // and map to Country, Title, NumberOfContracts
        JavaRDD<String> mostPopularTitlePerCountry = countryNumContractsTitle.join(countryMaxNull)
        .map(pair -> new String(pair._1()._1()+ "," + pair._2()._1() + "," + pair._1()._2()));

        // Save the result to the output folder
        mostPopularTitlePerCountry.saveAsTextFile(outputPath2);

        sc.close();
    }
}
